# vim:set ts=2 sw=2 sts=2 et:
from rzn.api import obj, Transform, Plugin, Logger, APIException, Layers, units
from rzn.api.common import triangulate
from .rasterise import rasterise
from collections import defaultdict
import sys, os, math
import zipfile

log = Logger(__name__)


class GerberLayer(object):
  def __init__(self):
    self.apertures = defaultdict(lambda: [])

def inch(v):
  return v/float(units.inch(1))

class GerberExport(object):
  def __init__(self, project, config):
    self.project = project
    self.config = config
    # configure layers (razen->gerber mapping)
    self.layermap = {}
    layerconf = self.config.get("layermap", {})
    for rznlayer, gbrlayer in layerconf.items():
      if hasattr(Layers, rznlayer):
        self.layermap[getattr(Layers, rznlayer)] = gbrlayer
    self.layers = defaultdict(lambda: GerberLayer())
    self.drills = defaultdict(lambda: [])
    self.minx = self.miny = sys.maxint
    self.maxx = self.maxy = -sys.maxint

  def _layer_filename(self, basename, gbrlayer):
    filenames = self.config.get("filenames", {})
    if gbrlayer in filenames:
      fn = filenames[gbrlayer]
    elif "*" in filenames:
      fn = filenames["*"]
    else:
      fn = "{basefn}_{layername}.gbr"
    return fn.format(basefn=basename,layername=gbrlayer)

  def _getbounds(self, pts):
    for i in range(len(pts)//2):
      x,y = pts[i*2+0], pts[i*2+1]
      if x < self.minx: self.minx = x
      if x > self.maxx: self.maxx = x
      if y < self.miny: self.miny = y
      if y > self.maxy: self.maxy = y

  def _layer(self, razenlayer):
    if razenlayer not in self.layermap:
      return self.layers["SilkTop"]
    return self.layers[self.layermap[razenlayer]]

  def _add_points(self, layer, ap, pts):
    self._layer(layer).apertures[ap].append(pts)

  def _add_stop_points(self, layer, ap, pts):
    if layer == Layers.CopperTop:
      self._add_points(Layers.StopTop, ap, pts)
    elif layer == Layers.CopperBottom:
      self._add_points(Layers.StopBottom, ap, pts)

  def _add_stencil_points(self, layer, ap, pts):
    if layer == Layers.CopperTop:
      self._add_points(Layers.StencilTop, ap, pts)
    elif layer == Layers.CopperBottom:
      self._add_points(Layers.StencilBottom, ap, pts)

  def _render_segment(self, e, transform, side):
    pts = transform.points(e.pta+e.ptb)
    self._getbounds(pts)
    self._add_points(e.getLayer(side), ("C", e.width), pts)

  def _render_arc(self, e, transform, side):
    pts = e.genSegmented()
    self._getbounds(pts)
    for i in range((len(pts)/2)-1):
      seg = pts[i*2:i*2+4]
      self._add_points(e.getLayer(side), ("C", e.width), seg)

  def _render_footprint(self, e, transform, side):
    t = transform.trans(*e.pos)
    fside = 1 if e.mirror else 0
    if e.mirror:
      t = t.flipx()
    t = t.rot(e.angle)
    self._render(e.epart.layout, t, fside, [] if e.showname else [e.nameText()])

  def _render_drill(self, e, transform, side):
    pts = transform.points(e.pos)
    if e.diameter > 0:
      self.drills[e.diameter].append(pts)

  def _render_pad(self, e, transform, side):
    pts = transform.points(e.pos)
    self._getbounds(pts)
    # drill
    if e.drill > 0:
      self.drills[e.drill].append(pts)
    # pad
    if e.shape == "circle":
      oap, ow, oh, opts = "C", e.width, e.height, pts
    elif e.shape == "octagon":
      ta = int(round(transform.angle()))
      w2,h2 = e.width/2, e.height/2
      bpts = []
      da = 2*math.pi / 8.0
      ofs = 2*math.pi / 16.0
      for i in range(8):
        bpts.append(math.cos(i*da+ofs)*w2)
        bpts.append(math.sin(i*da+ofs)*h2)
      txf = Transform.ident().trans(*pts).rot(ta)
      tpts = txf.points(bpts)
      oap, ow, oh, opts = "R", e.width, e.height, tpts
    elif e.shape == "rect" or e.shape == "rectangle":
      ta = int(round(transform.angle()))
      if (ta % 180) == 0:
        oap, ow, oh, opts = "R", e.width, e.height, pts
      elif (ta % 180) == 90:
        oap, ow, oh, opts = "R", e.height, e.width, pts
      else:
        rw2,rh2 = e.width/2, e.height/2
        rpts = [-rw2,-rh2, +rw2,-rh2, +rw2,+rh2, -rw2,+rh2]
        txf = Transform.ident().trans(*pts).rot(ta)
        tpts = txf.points(rpts)
        oap, ow, oh, opts = "R", e.width, e.height, tpts
    elif e.shape == "polygon":
      ta = int(round(transform.angle()))
      txf = Transform.ident().trans(*pts).rot(ta)
      tpts = txf.points(e.polygon)
      oap, ow, oh, opts = "R", e.width, e.height, tpts
    else:
      raise APIException("unhandled pad shape: %s" % e.shape)
    
    def stop_dims(top):
      if isinstance(e, obj.Via):
        v = e.stoptop if top else e.stopbot
        return (oap, v, v)
      return (oap, ow, oh)

    # add points for copper and stop layers
    layer = e.getLayer(side)
    self._add_points(layer, (oap, ow, oh), opts)
    if e.stop:
      self._add_stop_points(layer, stop_dims(True), opts)
    if (e.drill > 0):
      self._add_points(Layers.mirror(layer), (oap, ow, oh), opts)
      if e.stop:
        self._add_stop_points(Layers.mirror(layer), stop_dims(False), opts)
    if e.stencil:
      self._add_stencil_points(layer, (oap, ow, oh), opts)

  def _render_text(self, e, transform, side):
    size = e.properties["size"]
    fnt = e.getFont()
    txt = e.evaluatedValue()
    w,h = fnt.textSize(txt, size)
    mirror = side or e.properties.get("mirror", False)
    lines = fnt.genText(txt, size, e.getCurrentTransform(), mirror=mirror)
    fw = max(units.mm(0.1), e.properties.get("width", 0))
    for i in range(len(lines)/4):
      self._add_points(e.getLayer(side), ("C", fw), lines[i*4:(i*4)+4])

  def _render_polygon(self, e, transform, side):
    layer = e.getLayer(side)
    space = max(e.space, units.mm(0.05))
    raster_width = max(e.width, units.mm(0.25))
    clipped = e.clip(space+raster_width/2)
    # add outline segments for clipped polygon
    for poly in clipped:
      for i in range(len(poly)/2-1):
        p0x,p0y,p1x,p1y = poly[i*2:i*2+4]
        self._add_points(layer, ("C", raster_width), (p0x,p0y,p1x,p1y))
      p0x,p0y = poly[-2:]
      p1x,p1y = poly[:2]
      self._add_points(layer, ("C", raster_width), (p0x,p0y,p1x,p1y))
    # triangluate and rasterise polygon and add segments
    lines = rasterise(triangulate(clipped), raster_width)
    for i in range(len(lines)/4):
      self._add_points(layer, ("C", raster_width), lines[i*4:i*4+4])
    # handle thermals
    if e.thermal > 0:
      for t in e.thermalpoints:
        self._add_points(layer, ("C", e.thermal), t)

  def _render(self, layout, transform, side=0, ignores=[]):
    for e in layout:
      if e in ignores:
        continue
      if isinstance(e, obj.Segment):
        self._render_segment(e, transform, side)
      elif isinstance(e, obj.Arc):
        self._render_arc(e, transform, side)
      elif isinstance(e, obj.Footprint):
        self._render_footprint(e, transform, side)
      elif isinstance(e, obj.Drill):
        self._render_drill(e, transform, side)
      elif isinstance(e, obj.Pad):
        self._render_pad(e, transform, side)
      elif isinstance(e, obj.Text):
        self._render_text(e, transform, side)
      elif isinstance(e, obj.Polygon):
        self._render_polygon(e, transform, side)
      else:
        raise APIException("unknown object: %s" % e)

  def _genLayer(self, fn, layer):
    ofn = file(fn, "wb")
    def emitf(s):
      print >> ofn, "%s*" % s

    def emitp(s):
      print >> ofn, "%%%s*%%" % s

    def gpt(x, y, m):
      x = int(inch(x-self.minx)*10000.)
      y = int(inch(self.maxy-y)*10000.)
      emitf("X%06dY%06dD%02d" % (x,y,m))

    emitf("G75") # mutli quadrant mode
    emitf("G70") # inches
    emitp("OFA0B0") # offset 0,0
    emitp("FSLAX24Y24") # no leading zeros, 2.4 format
    emitp("IPPOS") # image polarity: positive
    emitp("LPD") # dark polarity
    aps = {}
    for i, ap in enumerate(layer.apertures):
      aps[ap] = i+10
      if ap[0] == "C":
        emitp("ADD%02d%s,%0.4f" % (i+10, ap[0], inch(ap[1])))
      elif ap[0] == "R":
        emitp("ADD%02d%s,%0.4fX%0.4f" % (i+10, ap[0], inch(ap[1]), inch(ap[2])))
      else:
        raise APIException("Unhandle aperture '%s'" % (ap[0]))

    for ap in layer.apertures:
      emitf("D%02d" % aps[ap])
      for pts in layer.apertures[ap]:
        if len(pts) == 2:
          # pad - flash an aperture
          gpt(pts[0], pts[1], 3)
        elif len(pts) == 4:
          # segment - move and draw
          gpt(pts[0], pts[1], 2)
          gpt(pts[2], pts[3], 1)
        elif len(pts) >= 6:
          # rotated pad - do a region
          emitf("G36")
          for i in range(len(pts)/2):
            gpt(pts[i*2+0], pts[i*2+1], 2 if i==0 else 1)
          gpt(pts[0], pts[1], 1)
          emitf("G37")
        else:
          raise APIException("unhandled aperture: %s" % pts)
    emitf("M02")
    return fn

  def _genDrills(self, fn):
    ofn = file(fn, "wb")
    def emit(s):
      print >> ofn, s
    
    def gpt(x, y):
      x = int(inch(x-self.minx)*10000.)
      y = int(inch(self.maxy-y)*10000.)
      emit("X%04dY%04d" % (x,y))

    emit("%")
    emit("M48") # start header
    emit("M72") # inches
    drillcodes = {}
    for i, drill in enumerate(self.drills):
      emit("T%02dC%0.4f" % (i+1, inch(drill)))
      drillcodes[drill] = i+1
    emit("%")
    for drill in self.drills:
      emit("T%02d" % drillcodes[drill])
      for hole in self.drills[drill]:
        gpt(*hole)
    emit("M30") # end
    return fn

  def compress(self, fn, files):
    zip = zipfile.ZipFile("%s.zip" % fn, "w", zipfile.ZIP_DEFLATED)
    rootlen = len(os.path.dirname(fn)) + 1
    for gfile in files:
      zip.write(gfile, gfile[rootlen:])
    zip.close()

  def save(self, fn):
    self._render(self.project.layout.objs(), Transform.ident())
    gerbers = []
    for lname, layer in self.layers.items():
      gerbers.append(self._genLayer(self._layer_filename(fn, lname), layer))
    gerbers.append(self._genDrills(self._layer_filename(fn, "Drill")))
    self.compress(fn , gerbers)


class GerberExportPlugin(Plugin):
  def __init__(self):
    Plugin.__init__(self, "Gerber", Plugin.EXPORT)

  def main(self, ui):
    self.batchMain(ui.currentTab().project(), {})

  def batchMain(self, proj, options):
    exp = GerberExport(proj, self.loadConfig("gerber"))
    exp.save(os.path.join(proj.filename(), os.path.basename(proj.filename())))
