# vim:set ts=2 sw=2 sts=2 et:
from bezier import bezier

BEZIER_PTS = 5

# Parse an SVG path string and generate a set of polygons
# http://www.w3.org/TR/SVG/paths.html
class PathParser(object):
  def __init__(self, s):
    self.s = s
    self.pos = 0

  # lexing
  
  def get(self):
    c = self.s[self.pos]
    self.pos += 1
    return c

  def peek(self):
    if self.pos < len(self.s):
      return self.s[self.pos]

  def eatnum(self):
    r = ""
    if self.peek() == "-":
      r += self.get()
    while self.peek() and self.peek() in "0123456789.":
      r += self.get()
    return float(r)

  def eatnums(self, n):
    r = []
    for i in range(n):
      r.append(self.eatnum())
      self.eatcomma()
    return r

  def eatcomma(self):
    if self.peek() == ",":
      self.get()
  
  def eatspace(self):
    while self.peek() == " ":
      self.get()

  def eatcmd(self):
    cmds = {
      "M" : (2, self.move_abs),
      "L" : (2, self.line_abs),  "l" : (2, self.line_rel),
      "H" : (1, self.hline_abs), "h" : (1, self.hline_rel),
      "V" : (1, self.vline_abs), "v" : (1, self.vline_rel),
      "C" : (6, self.curve_abs), "c" : (6, self.curve_rel),
      "S" : (4, self.smoothcurve_abs), "s" : (4, self.smoothcurve_rel),
      "z" : (0, self.closepath)
    }
    cmd = self.get()
    if cmd in cmds:
      argcount, fn = cmds[cmd]
      if argcount:
        fn(*self.eatnums(argcount))
      else:
        fn()
    else:
      raise Exception("unhandled path command %s" % (cmd))
    self.eatspace()

   # parser driver

  def parse(self):
    self.polys = []
    self.pts = []
    self.pt = (0,0)
    self.p1 = (0,0)
    while self.pos < len(self.s):
      self.eatcmd()
    if len(self.pts):
      self.polys.append(list(self.pts))
    return self.polys

  # polygon rendering

  def addpt(self, x, y):
    self.pt = (x, y)
    self.pts.extend(self.pt)

  def move_abs(self, x, y):
    self.closepath()
    self.addpt(x,y)
  
  def line_abs(self, x, y):
    self.addpt(x,y)
  
  def line_rel(self, dx, dy):
    self.addpt(self.pt[0]+dx, self.pt[1]+dy)
  
  def hline_abs(self, x):
    self.addpt(x, self.pt[1])
  
  def hline_rel(self, dx):
    self.addpt(self.pt[0]+dx, self.pt[1])
  
  def vline_abs(self, y):
    self.addpt(self.pt[0], y)
  
  def vline_rel(self, dy):
    self.addpt(self.pt[0], self.pt[1]+dy)
  
  def curve_abs(self, x1, y1, x2, y2, x, y):
    self.p1 = (x1,y1)
    self.pts.extend(bezier(self.pt, (x1, y1), (x2, y2), (x, y), BEZIER_PTS))
    self.pt = (x,y)
  
  def curve_rel(self, dx1, dy1, dx2, dy2, dx, dy):
    x, y = self.pt[0] + dx, self.pt[1] + dy
    x1, y1 = self.pt[0] + dx1, self.pt[1] + dy1
    x2, y2 = self.pt[0] + dx2, self.pt[1] + dy2
    self.curve_abs(x1, y1, x2, y2, x, y)
  
  def smoothcurve_abs(self, x2, y2, x, y):
    x1 = self.pt[0] + (self.pt[0] - self.p1[0])
    y1 = self.pt[1] + (self.pt[1] - self.p1[1])
    self.curve_abs(x1, y1, x2, y2, x, y)
  
  def smoothcurve_rel(self, dx2, dy2, dx, dy):
    x, y = self.pt[0] + dx, self.pt[1] + dy
    x2, y2 = self.pt[0] + dx2, self.pt[1] + dy2
    self.smoothcurve_abs(x2, y2, x, y)

  def closepath(self):
    if len(self.pts):
      self.polys.append(list(self.pts))
      self.pts = []

